; -*- fundamental -*-
; $Id: keyboard.asm,v 1.6 2004/09/16 01:08:32 hpa Exp $
; Keyboard controller program written in Z80 assembly
;
; The "status byte" emitted is:
; 0 - Shift
; 1 - Ctrl
; 2 - Alt
; 3 - Win
; 4 - Scroll Lock
; 5 - Num Lock
; 6 - Caps Lock
; 7 - strobe

rst0
	; Enter here after reset
	di
	ld sp,#2048
	jp (_start)
	nop

rst8
	; Print a character in A on serial port
	push af
sc_notready
	in a,(#1)
	cp #0x80		; Bit 7 = serial port ready
	jr c,sc_notready
	pop af
	out (#1),a
	ret

	nop
	nop
	nop
	nop
	nop

	
rst24
	; Print a string -> DE on serial port
	; (DE is used because it is the register pair least
	; likely to be actively in use)
	push af
r24_loop
	ld a,(de)
	and a
	jr z,r24_done
	rst 8
	inc de
	jr r24_loop
r24_done
	pop af
	ret


	;
	; Delay a specific number of microseconds in A; this
	; assumes a 25 MHz clock.  A = 0 on return
	;
udelay
	srl a			; 25 MHz - divide by 4
udelay_l
	push hl			; 11 cycles
	inc hl			; 6 cycles
	dec hl			; 6 cycles
	pop hl			; 10 cycles
	dec a			; 4 cycles
	jr nz,udelay_l		; 12 cycles (taken)
	ret

	;
	; Delay a specific number of 1/8 milliseconds
	; in A; this assumes a 25 MHz clock.  A = 0 on return
	;
omsdelay
	push bc
omsdelay_l1
	ld b,#239		; 7 cycles
omsdelay_l2
	djnz omsdelay_l2	; 13*239-5 cycles
	dec a			; 4 cycles
	jr nz,omsdelay_l1	; 12 cycles (taken)
	pop bc
	ret

	; Probe the status of the keyboard port; require N/2
	; samples to consider it debounced; should have the
	; mask to probe in E.  Destroys A.  C=1 on return if
	; the bit was set.
debounce
	push bc
;	ld bc,#0x1008		; Need 8/16 samples
	ld bc,#0x0402		; Need 2/4 samples
db_loop
	in a,(#0)
	and e
	jr z,db_zero
	dec c
db_zero
	djnz db_loop
	ld a,c
	pop bc
	add a,#0x80
	ret
		
	;
	; Read a byte from the keyboard -> A
	;
read_byte
	push bc
	push de
	push hl
	
	ld hl,#0		; Accumulated word
	ld b,#11		; Bits to read (including start+parity+stop)
	
	ld a,#0x1F		; Control port; no drivers
	out (#0),a
rb_loop
rb_waitclk			; Wait for CLK# = 0
	ld e,#0x02
	call (debounce)
	jr c,rb_waitclk

;	ld a,#2
;	call (udelay)

	ld e,#0x01
	call (debounce)
	rr h
	rr l

rb_waitnoclk			; Wait for CLK# = 1
	ld e,#0x02
	call (debounce)
	jr nc,rb_waitnoclk

	; More bits to get...
	djnz rb_loop

	; Now the proper value is in bits <13:6> in HL
	sla l
	rl h
	rl l	; Parity bit <- bit 0 of L
	rl h
	
	ld a,h
	pop hl
	pop de
	pop bc
	ret


	;
	; Send a byte (A) to the keyboard
	;
write_byte
	push bc
	push de
	push hl

	ld h,#0xfe		; Stop+ack bit
	and a
	jp po,(wb_parity)
	inc h			; If parity is not odd, set parity bit
wb_parity
	ld l,a
	; Now HL contains the whole bit string to be sent, including
	; data+parity+stop+ack, but not including the start bit,
	; starting with the LSB.  The start bit is a bit special,
	; since we have to clock it ourselves.  On the other hand,
	; ack is just a normal bit for which we do not assert the data
	; line.
	ld b,#11		; Numbers of bits in HL

	ld a,#0x1d		; Assert CLK#
	out (#0),a
	ld a,#100		; 100 us
	call (udelay)

	ld a,#0x1c		; Assert DATA#
	out (#0),a
	ld a,#10		; 10 us
	call (udelay)

	ld a,#0x1e		; Release CLK#
	out (#0),a

wb_loop
wb_waitclk			; Wait for CLK# = 0
	ld e,#0x02
	call (debounce)
	jr c,wb_waitclk

	ld a,#5
	call (udelay)
	; A = 0

	srl h
	rr l
	; Now flag C = bit to send
	adc a,#0x1e		; 1E for zero, 1F for one
	out (#0),a

wb_waitnoclk			; Wait for CLK# = 1
	ld e,#0x02
	call (debounce)
	jr nc,wb_waitnoclk

	; More to send...
	djnz wb_loop

	pop hl
	pop de
	pop bc
	ret
	
	;
	; Display a hexadecimal number on the serial port
	;
printhex
	push af
	push hl
	push de
	push af
	ld de,hextbl
	ld h,#0
	srl a
	srl a
	srl a
	srl a
	ld l,a
	add hl,de
	ld a,(hl)
	rst 8
	ld h,#0
	pop af
	and #0x0F
	ld l,a
	add hl,de
	ld a,(hl)
	rst 8
	ld a,#32
	rst 8
	pop de
	pop hl
	pop af
	ret
	
hextbl
	byte "0123456789ABCDEF"

	;
	; Send a byte to the keyboard that is expected to be replied
	; with ACK (FA).  If we get resend (FE) instead, resend,
	; otherwise assume it was a stray byte and wait.
	; Returns reply in A; trashes E
	;
send_with_ack
	ld e,a
	push af				; Debug
	ld a,#0x3c
	rst 8
	pop af
	call (printhex)
	
	call (write_byte)
swa_reply
	call (read_byte)
	push af				; Debug
	ld a,#0x3e
	rst 8
	pop af
	call (printhex)
	cp #0xfa
	ret z
	cp #0xfe
	ld a,e
	jr z,send_with_ack
	jr swa_reply
	
_start
	ld iy,variables			; Point to variable area

	; Test
	ld a,#0x69
	call (printhex)

kt_reset
	ld a,#0xff			; Reset keyboard
	call (send_with_ack)
	call (read_byte)		; Post-reset status code
	call (printhex)
	ld a,#0xed			; Set LEDs
	call (send_with_ack)
	ld a,(lock_mask)		; Match the current lock mask
	call (send_with_ack)
	ld a,#0xf0			; Set scan code set
	call (send_with_ack)
	ld a,#0x03
	call (send_with_ack)
	ld a,#0xf8			; Set all keys Make/Break
	call (send_with_ack)
	jp (set_status)			; Compute status byte
	
	;
	; Keyboard main loop - read a char, decode, and emit
	;
main_loop_clear
kt_ignore
	; Clear the break flag
	xor a
	ld (break_flag),a

main_loop
	call (read_byte)
	call (printhex)
	ld hl,keytype_table
	ld d,#0
	ld e,a				; DE <- keynum
	add hl,de
	ld ix,keydata_table
	add ix,de
	add ix,de
	add ix,de
	add ix,de			; IX <- kd_t+4*keynum
	; D=0 still
	ld e,(hl)			; DE = keytype
	ld hl,keytype_handlers
	add hl,de
	add hl,de
	add hl,de			; HL = kt_h+3*type
	ld bc,(break_flag)		; B <- status_byte, C <- break_flag
	bit 0,c				; Lots of things need this
	jp (hl)
;
; Entry condititions:
;  A = keynum
;  B = status_byte
;  C = break_flag
;  D = 0
;  E = type code
; HL = handler
; IX = keydata_table+4*keynum
; IY = variables
; ZF = set if break_flag=0

kt_break
	ld a,#0xFF
	ld (break_flag),a
	jp (main_loop)

kt_shift
	jr nz,kts_break
	
	ld a,(shift_l_mask)
	or (ix+0)
	ld (shift_l_mask),a
	ld a,(shift_r_mask)
	or (ix+1)
	ld (shift_r_mask),a
	jr set_status

kts_break
	ld a,(shift_l_mask)
	and (ix+2)
	ld (shift_l_mask),a
	ld a,(shift_r_mask)
	and (ix+3)
	ld (shift_r_mask),a
	jr set_status

kt_lock
	jr nz,main_loop_clear
	ld a,#0xed			; Set LEDs
	call (send_with_ack)
	ld a,(lock_mask)
	xor (ix+0)
	ld (lock_mask),a
	call (send_with_ack)
	; jr set_status

set_status
	ld de,(status_byte)		; E <- status_byte, D <- lock_mask
	ld a,e
	and #0x80
	sla d
	sla d
	sla d
	sla d
	or d
	ld de,(shift_l_mask)
	or e	; shift_l_mask
	or d	; shift_r_mask
	ld (status_byte),a
	out (#3),a
	jp (main_loop_clear)

kt_symbol
	jr nz,ktas_break
	ld (down_key),a
	ld a,b			; status_byte
ktas_common
	and #0x03
	ld e,a
	; D=0; now DE = shift mode

	; If the strobe is currently active, lower
	; the strobe for 1 ms *before* switching
	; to the new keysym
	bit 7,b
	jr z,ktas_strobeok
	ld a,b
	and #0x7F
	out (#3),a
	ld a,#8			; Give it 1 ms
	call (omsdelay)

ktas_strobeok
	add ix,de		; IX->keysym
	ld a,(ix+0)
	out (#2),a
	ld a,b
	or #0x80		; strobe
	ld (status_byte),a
	out (#3),a	
	jp (main_loop_clear)

ktas_break
	cp (iy+5)		; down_key
	jp nz,(main_loop_clear)
	ld a,b			; status_byte
	and #0x7F		; strobe
	ld (status_byte),a
	out (#3),a
	jp (main_loop_clear)

kt_alpha
	jr nz,ktas_break
	ld (down_key),a
	ld a,b			; status_byte
	and #0x42		; CapsLock+Ctrl
	cp #0x40		; CapsLock on, Ctrl off
	ld a,b			; status_byte
	jr nz,ktas_common
	xor #0x01		; Invert sense of shift
	jr ktas_common

kt_scroll
	jr nz,ktas_break
	ld (down_key),a
	ld a,b
	bit 4,a			; Scroll lock
	jr z,ktas_common
	xor #0x02		; Inverse sense of Ctrl
	jr ktas_common

;
; Handlers for the various entries
;
keytype_handlers
	jp (kt_ignore)		; 0
	jp (kt_reset)		; 1
	jp (kt_break)		; 2
	jp (kt_shift)		; 3
	jp (kt_lock)		; 4
	jp (kt_symbol)		; 5
	jp (kt_alpha)		; 6
	jp (kt_scroll)		; 7

;
; Status variables
;
variables				; IY points here
shift_l_mask				; (iy+0)
	byte 0
shift_r_mask				; (iy+1)
	byte 0
break_flag				; (iy+2)
	byte 0				; 0 for no break, ~0 for break
status_byte				; (iy+3)
	byte 0				; Current value for the status byte
lock_mask				; (iy+4)
	byte 02h			; Num Lock on
down_key				; (iy+5)
	byte 0				; Currently "down" key, if any

;
; Primary code table -- for each possible receive code classify
;                       it by type.  This is keyboard mode 3!
;
; Types are as follows:
; 0 - Ignore
; 1 - Reset (keyboard needs reinitialization)
; 2 - Break prefix
; 3 - Shift key
; 4 - Lock key
; 5 - Symbol key
; 6 - Alpha key (symbol key sensitive to caps lock)
; 7 - Arrow key (sensitive to scroll lock)
;
keytype_table
	byte 0,0,0,0,0,0,0,0, 5,0,0,0,0,5,5,0 ; 0x00
	byte 0,3,3,5,4,6,5,0, 0,3,6,6,6,6,5,0 ; 0x10
	byte 0,6,6,6,6,5,5,0, 0,5,6,6,6,6,5,0 ; 0x20
	byte 0,6,6,6,6,6,5,0, 0,3,6,6,6,5,5,0 ; 0x30
	byte 0,5,6,6,6,5,5,0, 0,5,5,6,6,6,5,0 ; 0x40
	byte 0,0,6,5,6,6,0,0, 3,3,5,6,0,0,0,4 ; 0x50
	byte 5,7,0,5,5,5,5,5, 0,5,7,5,5,5,5,5 ; 0x60
	byte 5,5,5,5,5,5,4,5, 0,5,5,0,5,5,5,0 ; 0x70
	byte 0,0,0,0,5,0,0,0, 0,0,0,3,3,0,0,0 ; 0x80
	byte 0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0 ; 0x90
	byte 0,0,0,0,0,0,0,0, 0,0,1,0,0,0,0,0 ; 0xA0
	byte 0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0 ; 0xB0
	byte 0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0 ; 0xC0
	byte 0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0 ; 0xD0
	byte 1,1,0,0,0,0,0,0, 0,0,0,0,0,0,0,0 ; 0xE0
	byte 2,0,0,0,0,0,0,0, 0,0,1,0,0,0,1,1 ; 0xF0

;
; Data table.  Each entry is four bytes; the meaning
; is as follows:
;
; Type	Byte 0	Byte 1	Byte 2	Byte 3
; ------------------------------------
; 0	-	-	-	-
; 1	-	-	-	-
; 2	-	-	-	-
; 3	LMask	RMask	~LMask	~RMask
; 4	Mask	-	-	-
; 5	Normsl	Shift	Ctrl	S+C
; 6	Normal	Caps	Ctrl	S+C
; 7	Normal	Normal	Scroll	Scroll
;
; This table only need to extend to cover the range
; that has types 3-7 in the keytype_table.
keydata_table
	byte 0x00,0x00,0x00,0x00 ; 0x00
	byte 0x00,0x00,0x00,0x00 ; 0x01
	byte 0x00,0x00,0x00,0x00 ; 0x02
	byte 0x00,0x00,0x00,0x00 ; 0x03
	byte 0x00,0x00,0x00,0x00 ; 0x04
	byte 0x00,0x00,0x00,0x00 ; 0x05
	byte 0x00,0x00,0x00,0x00 ; 0x06
	byte 0x00,0x00,0x00,0x00 ; 0x07
	byte 0x1B,0x1B,0x1B,0x1B ; 0x08
	byte 0x00,0x00,0x00,0x00 ; 0x09
	byte 0x00,0x00,0x00,0x00 ; 0x0A
	byte 0x00,0x00,0x00,0x00 ; 0x0B
	byte 0x00,0x00,0x00,0x00 ; 0x0C
	byte 0x09,0x09,0x09,0x09 ; 0x0D
	byte 0x7F,0x7F,0x1F,0x1F ; 0x0E
	byte 0x00,0x00,0x00,0x00 ; 0x0F
	byte 0x00,0x00,0x00,0x00 ; 0x10
	byte 0x02,0x00,0xFD,0xFF ; 0x11
	byte 0x01,0x00,0xFE,0xFF ; 0x12
	byte "<" ,">" ,0x7F,0x7F ; 0x13
	byte 0x04,0x00,0x00,0x00 ; 0x14
	byte "q" ,"Q" ,0x11,0x01 ; 0x15
	byte "1" ,0x21,"1" ,0x21 ; 0x16
	byte 0x00,0x00,0x00,0x00 ; 0x17
	byte 0x00,0x00,0x00,0x00 ; 0x18
	byte 0x04,0x00,0xFB,0xFF ; 0x19
	byte "z" ,"Z" ,0x1A,0x0A ; 0x1A
	byte "s" ,"S" ,0x13,0x03 ; 0x1B
	byte "a" ,"A" ,0x01,0x11 ; 0x1C
	byte "w" ,"W" ,0x17,0x07 ; 0x1D
	byte "2" ,0x22,"2" ,0x22 ; 0x1E
	byte 0x00,0x00,0x00,0x00 ; 0x1F
	byte 0x00,0x00,0x00,0x00 ; 0x20
	byte "c" ,"C" ,0x03,0x13 ; 0x21
	byte "x" ,"X" ,0x18,0x08 ; 0x22
	byte "d" ,"D" ,0x04,0x14 ; 0x23
	byte "e" ,"E" ,0x05,0x15 ; 0x24
	byte "4" ,"$" ,"4" ,"$"  ; 0x25
	byte "3" ,"#" ,"3" ,"#"  ; 0x26
	byte 0x00,0x00,0x00,0x00 ; 0x27
	byte 0x00,0x00,0x00,0x00 ; 0x28
	byte " " ," " ,0x00,0x00 ; 0x29
	byte "v" ,"V" ,0x16,0x06 ; 0x2A
	byte "f" ,"F" ,0x06,0x16 ; 0x2B
	byte "t" ,"T" ,0x14,0x04 ; 0x2C
	byte "r" ,"R" ,0x12,0x02 ; 0x2D
	byte "5" ,"%" ,"5" ,"%"  ; 0x2E
	byte 0x00,0x00,0x00,0x00 ; 0x2F
	byte 0x00,0x00,0x00,0x00 ; 0x30
	byte "n" ,"N" ,0x0E,0x1E ; 0x31
	byte "b" ,"B" ,0x02,0x12 ; 0x32
	byte "h", "H", 0x08,0x18 ; 0x33
	byte "g" ,"G" ,0x07,0x17 ; 0x34
	byte "y" ,"Y" ,0x19,0x09 ; 0x35
	byte "6" ,"&" ,"6" ,"&"  ; 0x36
	byte 0x00,0x00,0x00,0x00 ; 0x37
	byte 0x00,0x00,0x00,0x00 ; 0x38
	byte 0x00,0x04,0xFF,0xFB ; 0x39
	byte "m" ,"M" ,0x0D,0x1D ; 0x3A
	byte "j" ,"J" ,0x0A,0x1A ; 0x3B
	byte "u" ,"U" ,0x15,0x05 ; 0x3C
	byte "7" ,"/" ,"7" ,"/"  ; 0x3D
	byte "8" ,"(" ,"8" ,"("  ; 0x3E
	byte 0x00,0x00,0x00,0x00 ; 0x3F
	byte 0x00,0x00,0x00,0x00 ; 0x40
	byte "," ,0x3B,"," ,0x3B ; 0x41
	byte "k" ,"K" ,0x0B,0x1B ; 0x42
	byte "i" ,"I" ,0x09,0x19 ; 0x43
	byte "o" ,"O" ,0x0F,0x1F ; 0x44
	byte "0" ,"=" ,"0" ,"="  ; 0x45
	byte "9" ,")" ,"9" ,")"  ; 0x46
	byte 0x00,0x00,0x00,0x00 ; 0x47
	byte 0x00,0x00,0x00,0x00 ; 0x48
	byte "." ,":" ,"." ,":"  ; 0x49
	byte "-" ,"_" ,0x1F,0x1F ; 0x4A
	byte "l" ,"L" ,0x0C,0x1C ; 0x4B
	byte 0x7C,0x5C,0x1C,0x0C ; 0x4C
	byte "p" ,"P" ,0x10,0x00 ; 0x4D
	byte "+" ,"?" ,0x7F,0x7F ; 0x4E
	byte 0x00,0x00,0x00,0x00 ; 0x4F
	byte 0x00,0x00,0x00,0x00 ; 0x50
	byte 0x00,0x00,0x00,0x00 ; 0x51
	byte "{" ,"[" ,0x1B,0x0B ; 0x52
	byte "'" ,"*" ,"'" ,"*"  ; 0x53
	byte "}" ,"]" ,0x1D,0x0D ; 0x54
	byte "`", "@" ,0x00,0x10 ; 0x55
	byte 0x00,0x00,0x00,0x00 ; 0x56
	byte 0x00,0x00,0x00,0x00 ; 0x57
	byte 0x00,0x02,0xFF,0xFD ; 0x58
	byte 0x00,0x01,0xFF,0xFE ; 0x59
	byte 0x0D,0x0D,0x0D,0x0D ; 0x5A
	byte "~" ,"^" ,0x1E,0x0E ; 0x5B
	byte 0x00,0x00,0x00,0x00 ; 0x5C
	byte 0x00,0x00,0x00,0x00 ; 0x5D
	byte 0x00,0x00,0x00,0x00 ; 0x5E
	byte 0x01,0x00,0x00,0x00 ; 0x5F
	byte 0x1a,0x1a,0x1a,0x1a ; 0x60
	byte 0x08,0x08,0x01,0x01 ; 0x61
	byte 0x00,0x00,0x00,0x00 ; 0x62
	byte 0x17,0x17,0x17,0x17 ; 0x63
	byte 0x04,0x04,0x04,0x04 ; 0x64
	byte 0x1B,0x1B,0x1B,0x1B ; 0x65
	byte 0x08,0x08,0x08,0x08 ; 0x66
	byte 0x06,0x06,0x06,0x06 ; 0x67
	byte 0x00,0x00,0x00,0x00 ; 0x68
	byte "1" ,"1" ,"1" ,"1"  ; 0x69
	byte 0x09,0x09,0x13,0x13 ; 0x6A
	byte "4" ,"4" ,"4" ,"4"  ; 0x6B
	byte "7" ,"7" ,"7" ,"7"  ; 0x6C
	byte 0x1E,0x1E,0x1E,0x1E ; 0x6D
	byte 0x1C,0x1C,0x1C,0x1C ; 0x6E
	byte 0x1D,0x1D,0x1D,0x1D ; 0x6F
	byte "0" ,"0" ,"0" ,"0"  ; 0x70
	byte "." ,"." ,"." ,"."  ; 0x71
	byte "2" ,"2" ,"2" ,"2"  ; 0x72
	byte "5" ,"5" ,"5" ,"5"  ; 0x73
	byte "6" ,"6" ,"6" ,"6"  ; 0x74
	byte "8" ,"8" ,"8" ,"8"  ; 0x75
	byte 0x02,0x00,0x00,0x00 ; 0x76
	byte "/" ,"/" ,"/" ,"/"  ; 0x77
	byte 0x00,0x00,0x00,0x00 ; 0x78
	byte 0x0D,0x0D,0x0D,0x0D ; 0x79
	byte "3" ,"3" ,"3" ,"3"  ; 0x7A
	byte 0x00,0x00,0x00,0x00 ; 0x7B
	byte "+" ,"+" ,"+" ,"+"  ; 0x7C
	byte "9" ,"9" ,"9" ,"9"  ; 0x7D
	byte "*" ,"*" ,"*" ,"*"  ; 0x7E
	byte 0x00,0x00,0x00,0x00 ; 0x7F
	byte 0x00,0x00,0x00,0x00 ; 0x80
	byte 0x00,0x00,0x00,0x00 ; 0x81
	byte 0x00,0x00,0x00,0x00 ; 0x82
	byte 0x00,0x00,0x00,0x00 ; 0x83
	byte "-" ,"-" ,"-" ,"-"  ; 0x84
	byte 0x00,0x00,0x00,0x00 ; 0x85
	byte 0x00,0x00,0x00,0x00 ; 0x86
	byte 0x00,0x00,0x00,0x00 ; 0x87
	byte 0x00,0x00,0x00,0x00 ; 0x88
	byte 0x00,0x00,0x00,0x00 ; 0x89
	byte 0x00,0x00,0x00,0x00 ; 0x8A
	byte 0x08,0x00,0xF7,0xFF ; 0x8B
	byte 0x00,0x08,0xFF,0xF7 ; 0x8C
	byte 0x00,0x00,0x00,0x00 ; 0x8D
	byte 0x00,0x00,0x00,0x00 ; 0x8E
	byte 0x00,0x00,0x00,0x00 ; 0x8F
